import crypto
import os

import getpass

DIRECTORY = os.path.expanduser("~") + os.path.sep + ".jpass" + os.path.sep
SECRET_KEYFILE_PATH = DIRECTORY + ".keyfile"

def get_keyfile_data():
    """Prompts the user for their password, decrypts their keyfile and returns
    its contents -- effectively returns the encryption key for all of the
    encrypted entry files in ~/.jpass.
    """
    keyfile_data = ""
    while not keyfile_data:
        password = get_password(confirm_prompt="")
        try:
            keyfile_data = crypto.decrypt(password, SECRET_KEYFILE_PATH)
        except crypto.CryptoError:
            print "Failed to unlock keyfile; bad password?"
    return keyfile_data

def get_password(prompt="Password: ", confirm_prompt="Confirm: ",
                 err="Passwords did not match.\n"):
    """Prompts the user for a password.
        prompt         -- the prompt to show the user.
        confirm_prompt -- the confirmation prompt to show the user. If null, do
                          do not confirm password input.
        err            -- the error message to print if the initial password
                          input and the confirm password input were not equal.
    """
    password = getpass.getpass(prompt)


    if confirm_prompt:
        while getpass.getpass(confirm_prompt) != password:
            print err
            password = getpass.getpass(prompt)

    return password

def get_yes_no_response(prompt, err="Invalid response.\n", yes="y", no="n"):
    """Gets a yes/no response from the user. Returns True if the user said yes,
    false if the user said no.
        prompt -- the prompt to show the user.
        err    -- the error message to print if the user does not respond with
                  either of the strings given by the keyword arguments `yes` or
                  `no`.
        yes    -- if the user inputs this string, return True.
        no     -- if the user inputs this string, return False.
    """
    response = ""
    while 1:
        response = raw_input(prompt)
        if    response == yes: return True
        elif  response == no:  return False
        else: print err



try:
    from Crypto.Cipher import AES
except ImportError:
    print "jpass requires the pycrypto library to run."
    print "Please ensure that this is installed on your system."
    exit(1)

import hashlib
import os
import struct

class CryptoError(Exception):
    pass

def create_secret_keyfile(key, outfile, size=4096):
    """Creates the secret keyfile for the unified key setup.
       key     -- the encryption key used to encrypt the keyfile.
       size    -- the size of the keyfile in bytes (the greater, the better).
       outfile -- where to write the keyfile.
    """
    keyfile_data = os.urandom(size)
    encrypt(key, keyfile_data, outfile)

def raw_encrypt(key, data, chunk=1024*32):
    """Encrypts raw bytes.
        key   -- the encryption key to use.
        data  -- the data to encrypt.
        chunk -- the chunk size to use.
    """
    key = stretch_key(key)

    # Create random 16 byte initialization vector.  
    iv = os.urandom(16)
    size = len(data)

    # Using cipher block chain mode; effectively means each block of plaintext
    # is XORed with the previous block's ciphertext. This means each
    # ciphertext block is unique up to each point.  
    crypter = AES.new(key, AES.MODE_CBC, iv)
    encrypted = struct.pack("<Q", size) + iv + crypter.encrypt(SUCCESS_INDICATOR)

    for x in xrange(size // chunk + 1):
        td = data[x * chunk:(x + 1) * chunk]
        td = pad(td, 16, " ")
        encrypted += crypter.encrypt(td)

    return encrypted

def raw_decrypt(key, data, chunk=1024*32):
    """Decrypts raw bytes.
        key   -- the decryption key to attempt.
        data  -- the encrypted data to attempt to decrypt.
        chunk -- the chunk size to use.
    """
    key = stretch_key(key)
    retval = ""

    # Retrieve the original size and initialization vector.  
    origsize = struct.unpack("<Q", data[:struct.calcsize("<Q")])[0]
    data = data[struct.calcsize("Q"):]

    iv = data[:16]
    data = data[16:]

    crypter = AES.new(key, AES.MODE_CBC, iv)

    suc = crypter.decrypt(data[:len(SUCCESS_INDICATOR)])
    data = data[len(SUCCESS_INDICATOR):]
    
    # If this condition is true, we can assume with extremely high
    # probability that decryption was successful.  
    if suc != SUCCESS_INDICATOR:
        raise CryptoError("Decryption failed -- bad key or input data?")

    while 1:
        c = data[:chunk]
        data = data[chunk:]
        if not c:
            break
        retval += crypter.decrypt(c)
    
    return retval[:origsize]

def encrypt(key, data, outfile, file_permissions=0600):
    """Encrypts some data and writes it to a file 'outfile'.
        key     -- the encryption key to use.
        data    -- the data to encrypt.
        outfile -- the name of the file to output to.
    """
    with open(outfile, "wb") as f:
        # Write the initialization vector to the file.  
        f.write(raw_encrypt(key, data))

    # Assign file permissions.  
    os.chmod(outfile, file_permissions)

def decrypt(key, infile):
    """Decrypts some data from an 'infile' and returns the plaintext.
        key    -- the encryption key to use.
        infile -- the name of the file to take input from.
    """
    data = ""
    with open(infile, "rb") as f:
        data = f.read()
    
    return raw_decrypt(key, data)

def pad(data, bs, pad_char):
    """Pads a string. This is needed to transform a string of n bytes long
    to one whose length is equal to the specified `bs` argument.
        data     -- the data to pad.
        bs       -- the length of the new padded string (in bytes).
        pad_char -- the character to use for the padding
    """
    return data + (bs - len(data) % bs) * pad_char

def stretch_key(key, rounds=65536, salt="", algo=hashlib.sha256):
    """Stretches an encryption key by a default of 65536 rounds.
        key    -- the data to stretch.
        rounds -- how many iterations to spend rehashing the data.
        algo   -- the hash algorithm to use for the hashing.
    """
    stretched = key
    for _ in xrange(rounds):
        stretched = algo(stretched + key + salt).digest()
    return stretched

# This string is embedded in the 'header' of .jpass files, along with the file
# size and initialization vector (all of which are required for the encryption
# of entries).
#
# The purpose of the success indicator is as follows: when a .jpass file is
# decrypted, the header is examined for the success indicator string. Presence
# of the success indicator string means that we can assume, with extremely high
# probability, that decryption was successful. If the success indicator string
# is not found, then it means that decryption has failed. This is more than
# likely because the user has supplied an incorrect encryption key.
SUCCESS_INDICATOR = pad("SUCCESS", 16, " ")



import common
import crypto

from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
from collections import OrderedDict
import glob
import os
import pickle

class Entry(object):

    def __init__(self, name, **kwargs):
        self.name = name
        self.fields = OrderedDict(kwargs)
    
    def write(self, key, path):
        """Serializes the object in a string and writes it to a file in an
        encrypted form.
            key -- the encryption key to use.
            path -- where to write the file.
        """
        serialized = pickle.dumps(self)
        ciphertext = crypto.encrypt(key, serialized, path)

    def __str__(self):
        """Causes the str() function to return the entry details."""
        s = "== %s ==\n" % self.name

        for j, k in self.fields.iteritems():
            if k:
                s += "%s: %s\n" % (j, k)
        return s

def read(key, path):
    """Attempt to decrypt an entry file and deserialize it, loading the object
    back into memory so that its attributes can be read.
        key -- the decryption key to try and use.
        path -- where to read from.
    """
    plaintext = crypto.decrypt(key, path)
    obj = pickle.loads(plaintext)
    return obj

def init():
    """Perform initialisation tasks, specifically:
        1. If there is no ~/.jpass directory, create one.
        2. If the user doesn't have a keyfile, walk them through creating one.
    """

    # If the ~/.jpass directory doesn't exist, create it and make sure that only
    # the current user can read its contents.
    if not os.path.isdir(common.DIRECTORY):
        os.mkdir(common.DIRECTORY)
        os.chmod(common.DIRECTORY, 0700)

    # If the user doesn't have a secret keyfile yet, make one.
    if not os.path.isfile(common.SECRET_KEYFILE_PATH):
        print "We must first create a keyfile before you can start using jpass."
        print "Please enter a strong passphrase for the keyfile."
        key = common.get_password()
        
        crypto.create_secret_keyfile(key, common.SECRET_KEYFILE_PATH)

def change_passwd():
    """Provides an intuitive interface for the user to change their password."""
    print "First, please unlock your keyfile."
    keyfile_data = common.get_keyfile_data()

    print "\n",

    print "Enter a new password for your keyfile."
    key = common.get_password(prompt="New password: ")
    crypto.encrypt(key, keyfile_data, common.SECRET_KEYFILE_PATH)

    print "\n",

    print "Done!"
    print "You will now have to use your new password to access your entries."

def edit_entry(name):
    """Provides an intuitive interface for the user to edit existing entries."""
    path = common.DIRECTORY + name

    if os.path.isfile(path):

        key = common.get_keyfile_data()
        print "\n",

        entry = read(key, path)
        print "As it stands, %s contains the following data:\n" % path
        print entry
        
        while 1:
            print "Choose a field to edit (case sensitive; enter nothing when done):"
            i = raw_input("<username/password/url/notes/...> ")

            if not i: break
            entry.fields[i] = raw_input("%s: " % i)
            
            print "\n",
            print "%s now contains the following data:\n" % path
            print entry
        
        entry.write(key, path)
        print "\n",
        print "Done!"

    else:
        print "Entry %s doesn't exist, so you can't edit it." % path

def list_entries():
    """Lists all .jpass files in ~/.jpass."""
    entries = glob.glob(common.DIRECTORY + "*.jpass")

    if entries:
        print "You have the following entry files in ~/.jpass:"

        # Take all of the paths, take their suffixes and print them each on a
        # new line.
        print "\n".join(e.rpartition("/")[-1] for e in entries)

    else:
        print "You have no entry files in ~/.jpass."

def read_entry(name):
    """Attempts to decrypt and load a .jpass file and show the user its
    attributes.
    """
    path = common.DIRECTORY + name

    if os.path.isfile(path):

        key = common.get_keyfile_data()

        print "\n",
        print read(key, path)
        print "Done!"

    else:
        print "Entry %s doesn't exist, so you can't read it." % path

def remove_entry(name):
    """Removes an entry from ~/.jpass (if it exists)."""
    path = common.DIRECTORY + name

    if os.path.isfile(path):
        os.remove(path)
        print "Successfully removed %s." % path
        print "Done!"
    else:
        print "Entry %s doesn't exist, so you can't remove it." % path

def store_entry(name):
    """Provides an intuitive interface for the user to create new entries."""
    path = common.DIRECTORY + name

    if not os.path.isfile(path):

        key = common.get_keyfile_data()
        print "\n",
        username = raw_input("username: ")
        password = raw_input("password: ")
        url = raw_input("url: ")
        notes = raw_input("notes: ")
        print "\n",


        e = Entry(name)
        e.fields["username"] = username
        e.fields["password"] = password
        e.fields["url"] = url
        e.fields["notes"] = notes

        more_fields = common.get_yes_no_response(
            "Would you like to add more fields? <y/n> ",
            err = "Please say either y (yes) or n (no)."
        )

        if more_fields:
                
            even_more_fields = True
            while even_more_fields:
                add_new_field(e)
                print "\n",
                even_more_fields = common.get_yes_no_response(
                    "Would you like to add another field? <y/n>  ",
                    err = "Please say either y (yes) or n (no)."
                )

        e.write(key, path)
        print "Done!"

    else:
        print "%s is already an entry!" % path
        print "You can edit it or delete it using the edit or remove tasks."

def add_new_field(e):
    """Creates a prompt for the user to add a new field to an entry object."""
    new_field_name = raw_input("Enter a name for the new field: ")
    new_field_value = raw_input("Enter the value for the new field: ")
    e.fields[new_field_name] = new_field_value

def main():
    ap = ArgumentParser(description="Manage your passwords.",
                        formatter_class=ArgumentDefaultsHelpFormatter)
    ap.add_argument("task", help="The task to perform.")
    ap.add_argument("--entry", "-e", help="The entry to work on.")
    args = ap.parse_args()

    # This dictionary maps 'task' keywords to their corresponding functions.
    tasks = {

        "change" : change_passwd,
        "changepasswd" : change_passwd,
        "passwd" : change_passwd,

        "change" : edit_entry,
        "edit" : edit_entry,
        "modify" : edit_entry,
        "tweak" : edit_entry,

        "list" : list_entries,
        "ls" : list_entries,

        "look" : read_entry,
        "read" : read_entry,
        "see" : read_entry,
        "view" : read_entry,

        "delete" : remove_entry,
        "remove" : remove_entry,
        "rm" : remove_entry,

        "create" : store_entry,
        "make" : store_entry,
        "store" : store_entry,
    }

    init()

    requested_task = args.task.lower()
    try:
        tasks[requested_task](args.entry + ".jpass")
    except TypeError:
        # If the function doesn't want a name to work with,
        # don't give it one.
        # TODO: perhaps a better way to do this...  
        tasks[requested_task]()
    except KeyError:
        # If the task that the user has given doesn't exist, tell
        # them.  
        print "No such task '%s'." % requested_task

if __name__ == "__main__":
    try:
        main()
    except (EOFError, KeyboardInterrupt):
        print "\n",
        print "Interrupt caught; exiting. Have a nice day!"



from setuptools import *

kwargs = {
	"author" : "Joe Ellis",
	"author_email" : "joechrisellis@gmail.com",
	"description" : "A command line password manager.",
	"entry_points" : {"console_scripts" : ["jpass=jpass.jpass:main"]},
	"license" : "GPL v2",
	"name" : "jpass",
	"packages" : ["jpass"],
	"version" : "V1.1",
        "install_requires" : ["pycrypto"]
}

setup(**kwargs)
